# -*- coding: utf-8 -*-

"""
    colurful
    ~~~~~~~~

    Terminal string styling done right, in Python.

    :copyright: (c) 2017 by Timo Furrer <tuxtimo@gmail.com>
    :license: MIT, see LICENSE for more details.
"""

import os
import sys
import copy
import types
import platform

from contextlib import contextmanager

from .core import colurful
from . import terminal

#: Holds the current version
__version__ = '1.5.4'

# if we are on Windows we have to init colorama
if platform.system() == 'Windows':
    os.system('color')


class colurfulModule(types.ModuleType):
    """
    Module Type object for dynamic attribute access on
    this module.
    """
    def __init__(self, colurful, *args):
        super(colurfulModule, self).__init__(*args)
        self.colurful = colurful

    @contextmanager
    def with_setup(self, colormode=None, colorpalette=None, extend_colors=False):
        """
        Return a new colurful object with the given color config.
        """
        colurful = colurful(
            colormode=self.colurful.colormode,
            colorpalette=copy.copy(self.colurful.colorpalette)
        )

        colurful.setup(
            colormode=colormode, colorpalette=colorpalette, extend_colors=extend_colors
        )
        yield colurful

    @contextmanager
    def with_8_ansi_colors(self):
        yield colurful(
            colormode=terminal.ANSI_8_COLORS,
            colorpalette=copy.copy(self.colurful.colorpalette)
        )

    @contextmanager
    def with_16_ansi_colors(self):
        yield colurful(
            colormode=terminal.ANSI_16_COLORS,
            colorpalette=copy.copy(self.colurful.colorpalette)
        )

    @contextmanager
    def with_256_ansi_colors(self):
        yield colurful(
            colormode=terminal.ANSI_256_COLORS,
            colorpalette=copy.copy(self.colurful.colorpalette)
        )

    @contextmanager
    def with_true_colors(self):
        yield colurful(
            colormode=terminal.TRUE_COLORS,
            colorpalette=copy.copy(self.colurful.colorpalette)
        )

    @contextmanager
    def with_palette(self, colorpalette):
        yield colurful(
            colormode=self.colurful.colormode,
            colorpalette=colorpalette
        )

    @contextmanager
    def with_updated_palette(self, colorpalette):
        colurful = colurful(
            colormode=self.colurful.colormode,
            colorpalette=copy.copy(self.colurful.colorpalette),
        )
        colurful.update_palette(colorpalette)
        yield colurful

    @contextmanager
    def with_style(self, style_name):
        colurful = colurful(
            colormode=self.colurful.colormode,
            colorpalette={},
        )
        colurful.use_style(style_name)
        yield colurful

    def __getattr__(self, name):
        """
        Dynamically get methods from colurful object.
        """
        # check if original module had the requested attribute
        # if yes, we have to return it since we don't want to
        # break the module functionality
        orig_module = __name__ + '_orig'
        if orig_module in sys.modules:
            try:
                return getattr(sys.modules[orig_module], name)
            except AttributeError:
                pass  # fallback to colurful functionality

        return getattr(self.colurful, name)


if os.environ.get('colurful_NO_MODULE_OVERWRITE', '0') != '1':
    # Unfortunately, prior to Python 3.4 the replaced
    # modules gets garbage collected which will
    # destroy all currently imported modules.
    # To prevent this, is to store a backup of the original
    # modules in sys.modules.
    sys.modules[__name__ + '_orig'] = sys.modules[__name__]

    # expose the colurful class as module
    sys.modules[__name__] = colurfulModule(colurful(), __name__)

